from typing import Final

LARGEST_ANIMAL_SIZE: Final = 3.125
STICK_LENGTH: Final = 4.0
FOOD_TREE_BASE_HEIGHT = 2.0
CANONICAL_FOOD_HEIGHT_ON_TREE = 2.0
FOOD_TREE_VISIBLE_HEIGHT = 11.0
TREE_FOOD_OFFSET = 2.0
