extends SpecBase

class_name RecordingOptionsSpec

var user_id: String  #optional
var apk_version: String  #optional
var recorder_host: String  #optional
var recorder_port: int  #optional
var is_teleporter_enabled: bool
var resolution_x: int
var resolution_y: int
var is_recording_human_actions: bool
var is_remote_recording_enabled: bool
var is_adding_debugging_views: bool
var is_debugging_output_requested: bool
